package com.ub.misc.epak;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.Date;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.Signature;
import java.security.interfaces.RSAPublicKey;
import java.security.cert.X509Certificate;
import javax.security.auth.x500.X500Principal;

import org.bouncycastle.x509.X509V1CertificateGenerator;
import org.bouncycastle.openssl.PEMWriter;
import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.util.encoders.Base64;

/** eLink data signer, signature verifier and key pair generator.
 * @author SEB banka, Aldis Egle, 2008
 */
public class EPakSigner {

  private X509Certificate cert;
  private KeyPair kp;
  private Signature signer;
  private PublicKey pubKey;

  public EPakSigner() {
    // Registering BouncyCastle provider
    BouncyCastleProvider provider = new BouncyCastleProvider();
    Security.addProvider(provider);
  }

  /** Generates RSA public and secure key pair.
   * Public key is stored in self signed X.509 certificate.
   * Private key can be retrieved with <code>getPrivateKey</code>.
   * @param cn Common name (CN)
   * @param ou Organization unit (OU)
   * @param o Organization (O)
   * @param c Country (C)
   * @return X.509 certificate in PEM format.
   * @see #getPrivateKey
   */
  public String generatePair(String cn, String ou, String o, String c, int yearsValid) throws Exception {
    // Generates key pair
    KeyPairGenerator gen = KeyPairGenerator.getInstance("RSA");
    kp = gen.generateKeyPair();
    // Generates certificate
    X509V1CertificateGenerator certGen = new X509V1CertificateGenerator();
    X500Principal issuer = new X500Principal("CN="+cn+",OU="+ou+",O="+o+",C="+c);
    certGen.setPublicKey(kp.getPublic());
    certGen.setSerialNumber(((RSAPublicKey)kp.getPublic()).getModulus());
    certGen.setNotBefore(new Date());
    certGen.setNotAfter(new Date(new Date().getTime() + 31536000000L * yearsValid));
    certGen.setSubjectDN(issuer);
    certGen.setIssuerDN(issuer);
    certGen.setSignatureAlgorithm("SHA1withRSA");
    X509Certificate cert = certGen.generate(kp.getPrivate());
    // Encoding to Base64
    StringWriter strw = new StringWriter();
    PEMWriter pemw = new PEMWriter(strw);
    pemw.writeObject(cert);
    pemw.close();
    // Returning certificate
    return strw.toString();
  }

  /** After generating a key pair this returns secure key encoded in Base64 format.
   * @return Base64 encoded secure key.
   * @see #generatePair
   */
  public String getPrivateKey() throws Exception {
    // Encoding to Base64
    StringWriter strw = new StringWriter();
    PEMWriter pemw = new PEMWriter(strw);
    pemw.writeObject(kp.getPrivate());
    pemw.close();
    // Returning private key
    return strw.toString();
  }

  /** Initializes data signer with given secure key.
   * @param Base64 encoded secure key.
   * @see #add
   * @see #endSign
   */
  public void beginSign(String secKeyData) throws Exception {
    // Getting private key
    PEMReader pemr = new PEMReader(new StringReader(secKeyData));
    PrivateKey secKey = ((KeyPair)pemr.readObject()).getPrivate();
    // Creting and initializing signer component
    signer = Signature.getInstance("SHA1withRSA", BouncyCastleProvider.PROVIDER_NAME);
    signer.initSign(secKey);
  }

  /** Initializes signature verifier with given X.509 certificate.
   * @param certPem PEM encoded X.509 certificate containing RSA public key.
   * @see #add
   * @see #endVerify
   */
  public void beginVerify(String certPem) throws Exception {
    try {
      // Decoding certificate and getting public key
      PEMReader pemr = new PEMReader(new StringReader(certPem));
      cert = (X509Certificate)pemr.readObject();
      pubKey = cert.getPublicKey();
      // Creting and initializing signer component
      signer = Signature.getInstance("SHA1withRSA", BouncyCastleProvider.PROVIDER_NAME);
      signer.initVerify(pubKey);
    } catch (Exception ex) {
      cert = null;
      pubKey = null;
      throw ex;
    }
  }

  /** Adds string to signer component for data signing or signature validation.
   * Before given string is added 3 digit string length value.
   * @param value String value to add to signature component.
   */
  public void add(String value) throws Exception {
    // Adding 3 digit sting length value
    String len = "" + value.length();
    while (len.length() < 3) len = "0" + len;
    // Adding string to signature component
    signer.update((len + value).getBytes("UTF-8"));
  }

  /** Creates Base64 encoded signature from added String values.
   * @return Base64 encoded signature.
   * @see #beginSign
   */
  public String endSign() throws Exception {
    byte[] sig = signer.sign();
    return new String(Base64.encode(sig));
  }

  /** Verifies if added string values compares to given signature.
   * @param signature Base64 encoded signature.
   * @return true - signature is valid, false - signature is not valid.
   */
  public boolean endVerify(String signature) throws Exception {
    if (pubKey == null) return false;
    return signer.verify(Base64.decode(signature));
  }

}
